import asyncio
import json
import logging
import time

from pyweatherflowudp.aioudp import open_local_endpoint
from pyweatherflowudp.const import (
    DEFAULT_HOST,
    DEFAULT_PORT,
    DEVICE_UPDATE_INTERVAL_SECONDS,
    EVENT_AIR_DATA,
    EVENT_HUB_STATUS,
    EVENT_RAPID_WIND,
    EVENT_SKY_DATA,
    EVENT_TEMPEST_DATA,
    PROCESSED_EVENT_EMPTY,
    SOCKET_CHECK_INTERVAL_SECONDS,
    UNIT_SYSTEM_METRIC,
)
from pyweatherflowudp.data import (
    WeatherflowStationStateMachine,
    station_update_from_udp_frames,
)

_LOGGER = logging.getLogger(__name__)


class WeatherFlowListner:
    """Updates device states and attributes."""

    def __init__(
        self, station_id: str, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT, unit_system: str = UNIT_SYSTEM_METRIC
    ):
        self._station_id = station_id
        self._host = host
        self._port = port
        self._unit_system = unit_system

        self._last_device_update_time = 0
        self._last_udpsocket_check = 0
        self._station_state_machine = WeatherflowStationStateMachine()

        self._processed_data = {}
        self.last_update_id = None

        self.udp_connection = None
        self.udp_task = None
        self._udp_subscriptions = []
        self._is_first_update = True
        self._running = True

    async def update(self, force_update=False) -> dict:
        """Updates the status of devices."""

        if self._is_first_update:
            await self._get_device_list()

        await self.async_connect_socket()

        if self.udp_connection:
            _LOGGER.debug("Return data if updated by socket.")
            return self._processed_data

    async def async_connect_socket(self):
        """Connect the UDP Socket."""
        if self.udp_connection is not None:
            return

        if self.udp_task is not None:
            try:
                self.udp_task.cancel()
                self.udp_connection = None
            except Exception:
                _LOGGER.exception("Could not cancel udp_task")
        self.udp_task = asyncio.ensure_future(self._setup_socketreader())

    async def async_disconnect_socket(self):
        """Disconnect the websocket."""
        if self.udp_connection is None:
            return
        self._running = False
        await asyncio.sleep(2)

    async def _get_device_list(self) -> None:
        """Get Initialize the data array."""
        if self._is_first_update:
            self._update_station(self._station_id, PROCESSED_EVENT_EMPTY)
            self._station_state_machine.add(self._station_id, PROCESSED_EVENT_EMPTY)
        self._is_first_update = False

    async def _setup_socketreader(self):
        """Setup the UDP Socket Listner."""
        self.udp_connection = await open_local_endpoint(
            host=self._host, port=self._port
        )

        try:
            while self._running and not self.udp_connection.closed:
                try:
                    data, (host, port) = await self.udp_connection.receive()
                    if data is not None:
                        json_response = json.loads(data.decode("utf-8"))
                        self._process_message(json_response)
                    else:
                        break
                except Exception:  # pylint: disable=broad-except
                    _LOGGER.exception("Error processing websocket message")
                    return
        finally:
            _LOGGER.debug("Closing connection")
            await self.udp_connection.drain()
            self.udp_connection.close()
            self.udp_connection = None

    def _process_message(self, msg):
        """Process the UDP message."""
        msg_type = msg.get("type")
        if msg_type is not None:
            data_json = {}
            try:
                if msg_type in EVENT_HUB_STATUS:
                    data_json = {
                        "event_type": msg_type,
                        "hub_firmware_revision": msg["firmware_revision"],
                        "hub_uptime": msg["uptime"],
                        "hub_rssi": msg["rssi"],
                    }
                if msg_type in EVENT_RAPID_WIND:
                    obs = msg.get("ob")
                    data_json = {
                        "event_type": msg_type,
                        "time_epoch_rapid_wind": obs[0],
                        "wind_speed": obs[1],
                        "wind_bearing": obs[2],
                    }
                if msg_type in EVENT_SKY_DATA:
                    obs = msg["obs"][0]
                    data_json = {
                        "event_type": msg_type,
                        "time_epoch_sky": obs[0],
                        "illuminance": obs[1],
                        "uv": obs[2],
                        "rain_accumulated": obs[3],
                        "wind_lull": obs[4],
                        "wind_avg": obs[5],
                        "wind_gust": obs[6],
                        "wind_bearing": obs[7],
                        "battery_sky": obs[8],
                        "solar_radiation": obs[10],
                        "local_day_rain_accumulation": obs[11],
                        "precipitation_type": obs[12],
                    }
                if msg_type in EVENT_AIR_DATA:
                    obs = msg["obs"][0]
                    data_json = {
                        "event_type": msg_type,
                        "time_epoch_air": obs[0],
                        "station_pressure": obs[1],
                        "air_temperature": obs[2],
                        "relative_humidity": obs[3],
                        "lightning_strike_count": obs[4],
                        "lightning_strike_avg_distance": obs[5],
                        "battery_air": obs[6],
                    }
                if msg_type in EVENT_TEMPEST_DATA:
                    obs = msg["obs"][0]
                    data_json = {
                        "event_type": msg_type,
                        "time_epoch_tempest": obs[0],
                        "wind_lull": obs[1],
                        "wind_avg": obs[2],
                        "wind_gust": obs[3],
                        "wind_bearing": obs[4],
                        "station_pressure": obs[6],
                        "air_temperature": obs[7],
                        "relative_humidity": obs[8],
                        "illuminance": obs[9],
                        "uv": obs[10],
                        "solar_radiation": obs[11],
                        "local_day_rain_accumulation": obs[12],
                        "precipitation_type": obs[13],
                        "lightning_strike_count": obs[14],
                        "lightning_strike_avg_distance": obs[15],
                        "battery_tempest": obs[16],
                    }
            except Exception as err:
                _LOGGER.debug("Error occured processing data: %s", err)
                return

            # _LOGGER.debug("PROCESSED DATA: %s", data_json)

            if len(data_json) > 0:
                station_id, processed_station = station_update_from_udp_frames(
                    self._station_state_machine, self._station_id, self._unit_system, data_json
                )
                if station_id is None:
                    return

                processed_station.update(data_json)
                self.fire_event(self._station_id, processed_station)
            else:
                _LOGGER.debug("IGNORING TYPE: %s", msg_type)

    def subscribe_udpsocket(self, ws_callback):
        """Subscribe to udpsockets events.

        Returns a callback that will unsubscribe.
        """

        def _unsub_callback():
            self._udp_subscriptions.remove(ws_callback)

        _LOGGER.debug("Adding subscription: %s", ws_callback)
        self._udp_subscriptions.append(ws_callback)
        return _unsub_callback

    def fire_event(self, station_id, processed_event):
        """Callback and event to the subscribers and update data."""
        self._update_station(station_id, processed_event)

        for subscriber in self._udp_subscriptions:
            subscriber({station_id: self._processed_data[station_id]})

    def _update_station(self, station_id, processed_update):
        """Update internal state of a station."""
        self._processed_data.setdefault(station_id, {}).update(processed_update)



"""Constant definitions for SecSpy Wrapper."""

DEVICE_UPDATE_INTERVAL_SECONDS = 60
SOCKET_CHECK_INTERVAL_SECONDS = 120

DEFAULT_HOST = "0.0.0.0"
DEFAULT_PORT = 50222

EVENT_RAPID_WIND = "rapid_wind"
EVENT_HUB_STATUS = "hub_status"
EVENT_DEVICE_STATUS = "device_status"
EVENT_AIR_DATA = "obs_air"
EVENT_SKY_DATA = "obs_sky"
EVENT_TEMPEST_DATA = "obs_st"
EVENT_PRECIP_START = "evt_precip"
EVENT_STRIKE = "evt_strike"
SUPPORTED_EVENTS = [
    EVENT_DEVICE_STATUS,
    EVENT_HUB_STATUS,
    EVENT_RAPID_WIND,
    EVENT_AIR_DATA,
    EVENT_SKY_DATA,
    EVENT_TEMPEST_DATA,
    EVENT_STRIKE,
    EVENT_PRECIP_START,
]

PROCESSED_EVENT_EMPTY = {
    "time_epoch_rapid_wind": None,
    "time_epoch_sky": None,
    "time_epoch_air": None,
    "time_epoch_tempest": None,
    "wind_speed": 0,
    "wind_bearing": 0,
    "wind_direction": None,
    "station_pressure": 0,
    "air_temperature": 0,
    "relative_humidity": 0,
    "lightning_strike_count": 0,
    "lightning_strike_avg_distance": 0,
    "battery_air": 0,
    "battery_sky": 0,
    "battery_tempest": 0,
    "illuminance": 0,
    "uv": 0,
    "rain_accumulated": 0,
    "wind_lull": 0,
    "wind_avg": 0,
    "wind_gust": 0,
    "solar_radiation": 0,
    "local_day_rain_accumulation": 0,
    "precipitation_type": 0,
    "hub_firmware_revision": None,
    "hub_uptime": None,
    "hub_rssi": 0,
    "dewpoint": 0,
    "heatindex": 0,
}

UNIT_SYSTEM_METRIC = "metric"
UNIT_SYSTEM_IMPERIAL = "imperial"



""".WeatherFlow UDP Data."""
import logging
import math

from pyweatherflowudp.const import (
    EVENT_AIR_DATA,
    EVENT_HUB_STATUS,
    EVENT_RAPID_WIND,
    EVENT_SKY_DATA,
    EVENT_TEMPEST_DATA,
    UNIT_SYSTEM_METRIC,
)
_LOGGER = logging.getLogger(__name__)


def station_update_from_udp_frames(state_machine, station_id, unit_system, data_json):
    """Convert a websocket frame to internal format."""

    if not state_machine.has_device(station_id):
        _LOGGER.debug("Skipping non-configured station: %s", data_json)
        return None, None

    station = state_machine.update(station_id, data_json)

    # _LOGGER.debug("Processing station: %s", station)
    processed_station = process_station(station, unit_system)

    return station_id, processed_station


def process_station(station_data, unit_system):
    """Process the station json."""

    cnv = WeatherFunctions(unit_system)
    station_update = {}

    if station_data["event_type"] in EVENT_RAPID_WIND:
        station_update = {
            "time_epoch_rapid_wind": station_data["time_epoch_rapid_wind"],
            "wind_speed": cnv.windspeed(station_data["wind_speed"]),
            "wind_bearing": station_data["wind_bearing"],
            "wind_direction": cnv.winddirection_string(station_data["wind_bearing"]),
        }
    if station_data["event_type"] in EVENT_HUB_STATUS:
        station_update = {
            "hub_firmware_revision": station_data["hub_firmware_revision"],
            "hub_uptime": station_data["hub_uptime"],
            "hub_rssi": station_data["hub_rssi"],
        }
    if station_data["event_type"] in EVENT_SKY_DATA:
        station_update = {
            "time_epoch_sky": station_data["time_epoch_sky"],
            "illuminance": station_data["illuminance"],
            "uv": station_data["uv"],
            "rain_accumulated": cnv.rainvolume(station_data["rain_accumulated"]),
            "wind_avg": cnv.windspeed(station_data["wind_avg"]),
            "wind_gust": cnv.windspeed(station_data["wind_gust"]),
            "wind_lull": cnv.windspeed(station_data["wind_lull"]),
            "wind_bearing": station_data["wind_bearing"],
            "wind_direction": cnv.winddirection_string(station_data["wind_bearing"]),
            "solar_radiation": station_data["solar_radiation"],
            "local_day_rain_accumulation": cnv.rainvolume(station_data["local_day_rain_accumulation"]),
            "precipitation_type": station_data["precipitation_type"],
            "battery_sky": station_data["battery_sky"],
        }
    if station_data["event_type"] in EVENT_AIR_DATA:
        station_update = {
            "time_epoch_air": station_data["time_epoch_air"],
            "station_pressure": cnv.pressure(station_data["station_pressure"]),
            "air_temperature": station_data["air_temperature"],
            "relative_humidity": station_data["relative_humidity"],
            "lightning_strike_count": station_data["lightning_strike_count"],
            "lightning_strike_avg_distance": cnv.distance(station_data[
                "lightning_strike_avg_distance"
            ]),
            "battery_air": station_data["battery_air"],
            "dewpoint": cnv.dewpoint(station_data["air_temperature"], station_data["relative_humidity"]),
            "heatindex": cnv.heatindex(station_data["air_temperature"], station_data["relative_humidity"]),
        }
    if station_data["event_type"] in EVENT_TEMPEST_DATA:
        station_update = {
            "time_epoch_tempest": station_data["time_epoch_tempest"],
            "station_pressure": cnv.pressure(station_data["station_pressure"]),
            "air_temperature": station_data["air_temperature"],
            "relative_humidity": station_data["relative_humidity"],
            "lightning_strike_count": station_data["lightning_strike_count"],
            "lightning_strike_avg_distance": cnv.distance(station_data[
                "lightning_strike_avg_distance"
            ]),
            "illuminance": station_data["illuminance"],
            "uv": station_data["uv"],
            "wind_avg": cnv.windspeed(station_data["wind_avg"]),
            "wind_gust": cnv.windspeed(station_data["wind_gust"]),
            "wind_lull": cnv.windspeed(station_data["wind_lull"]),
            "wind_bearing": station_data["wind_bearing"],
            "wind_direction": cnv.winddirection_string(station_data["wind_bearing"]),
            "solar_radiation": station_data["solar_radiation"],
            "local_day_rain_accumulation": cnv.rainvolume(station_data["local_day_rain_accumulation"]),
            "precipitation_type": station_data["precipitation_type"],
            "battery_sky": station_data["battery_sky"],
            "battery_tempest": station_data["battery_tempest"],
            "dewpoint": cnv.dewpoint(station_data["air_temperature"], station_data["relative_humidity"]),
            "heatindex": cnv.heatindex(station_data["air_temperature"], station_data["relative_humidity"]),
        }

    return station_update

class WeatherFunctions:
    """Contains a set of Weather Formulas."""

    def __init__(self, unit_system):
        self._unit_system = unit_system

    def heatindex(self, temperature, humidity):
        """Returns the calculated Heat Index in Celcius."""
        T = temperature * 9/5 + 32 #Convert to Fahrenheit
        RH = humidity
        c1 = -42.379
        c2 = 2.04901523
        c3 = 10.14333127
        c4 = -0.22475541
        c5 = -6.83783e-3
        c6 = -5.481717e-2
        c7 = 1.22874e-3
        c8 = 8.5282e-4
        c9 = -1.99e-6

        # try simplified formula first (used for HI < 80)
        HI = 0.5 * (T + 61. + (T - 68.) * 1.2 + RH * 0.094)

        if HI >= 80:
            # use Rothfusz regression
            HI = math.fsum([
                c1,
                c2 * T,
                c3 * RH,
                c4 * T * RH,
                c5 * T**2,
                c6 * RH**2,
                c7 * T**2 * RH,
                c8 * T * RH**2,
                c9 * T**2 * RH**2,
            ])

        return round((HI - 32) * 5/9, 1)

    def dewpoint(self, temperature, humidity):
        """ Returns Dew Point in Celcius """
        return round(243.04*(math.log(humidity/100)+((17.625*temperature)/(243.04+temperature)))/(17.625-math.log(humidity/100)-((17.625*temperature)/(243.04+temperature))),1)

    def windchill(self, wind_speed, temperature):
        """ Returns Wind Chill in Celcius """
        if wind_speed < 1.3:
            return round(temperature,1)

        windKmh = wind_speed * 3.6
        return round(13.12 + (0.6215 * temperature) - (11.37 * math.pow(windKmh, 0.16)) + (0.3965 * temperature * math.pow(windKmh, 0.16)), 1)

    def feelslike(self, temperature, wind_chill, heat_index):
        """ Returns the Feels Like Temperature in Celcius """
        if temperature > 26.666666667:
            return heat_index
        elif temperature < 10:
            return wind_chill
        else:
            return round(temperature,1)

    def distance(self, value):
        """Return value based on unit system."""
        if self._unit_system in UNIT_SYSTEM_METRIC:
            return value

        return round(value * 0.621371192, 1)

    def pressure(self, value):
        """Return value based on unit system."""
        if self._unit_system in UNIT_SYSTEM_METRIC:
            return value

        return round(value * 0.0295299801647, 3)

    def rainvolume(self, value):
        """Return value based on unit system."""
        if value is None:
            return None

        if self._unit_system in UNIT_SYSTEM_METRIC:
            return value

        return round(value * 0.0393700787, 1)

    def windspeed(self, value):
        """Return value based on unit system."""
        if self._unit_system in UNIT_SYSTEM_METRIC:
            return value

        return round(value * 2.2369362921, 1)

    def winddirection_string(self, bearing):
        """Return a Wind Direction String from a degree bearing."""
        direction_array = ['N','NNE','NE','ENE','E','ESE','SE','SSE','S','SSW','SW','WSW','W','WNW','NW','NNW','N']
        direction = direction_array[int((bearing + 11.25) / 22.5)]
        return direction

class WeatherflowStationStateMachine:
    """A simple state machine for events."""

    def __init__(self):
        """Init the state machine."""
        self._stations = {}

    def has_device(self, station_id):
        """Check to see if a device id is in the state machine."""
        return station_id in self._stations

    def add(self, station_id, new_json):
        """Add a station in the state machine."""
        self._stations[station_id] = new_json

    def update(self, station_id, new_json):
        """Update an device in the state machine."""
        station_json = self._stations.get(station_id)
        if station_json is None:
            return None
        station_json.update(new_json)
        return station_json




"""Define package errors."""


class WeatherflowUdpError(Exception):
    """Define a base error."""


class RequestError(WeatherflowUdpError):
    """Define an error related to invalid requests."""


class ResultError(WeatherflowUdpError):
    """Define an error related to the result returned from a request."""
